# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Public API for tf.random namespace.
"""

from __future__ import print_function

from tensorflow.python import get_seed
from tensorflow.python import multinomial
from tensorflow.python import random_gamma as gamma
from tensorflow.python import random_normal as normal
from tensorflow.python import random_poisson as poisson
from tensorflow.python import random_shuffle as shuffle
from tensorflow.python import random_uniform as uniform
from tensorflow.python import set_random_seed
from tensorflow.python import truncated_normal
from tensorflow.python.ops.candidate_sampling_ops import log_uniform_candidate_sampler
from tensorflow.python.ops.candidate_sampling_ops import uniform_candidate_sampler

del print_function
