# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Keras built-in datasets.
"""

from __future__ import print_function

from tensorflow._api.v1.keras.datasets import boston_housing
from tensorflow._api.v1.keras.datasets import cifar10
from tensorflow._api.v1.keras.datasets import cifar100
from tensorflow._api.v1.keras.datasets import fashion_mnist
from tensorflow._api.v1.keras.datasets import imdb
from tensorflow._api.v1.keras.datasets import mnist
from tensorflow._api.v1.keras.datasets import reuters

del print_function
